#!/usr/bin/env python
#
# pycdda/drvr/linux.py
#
# This Python code is based heavily on the "C"-language implementation
# in "cd.c" from grip-2.95 by Mike Oliphant.
#
# Copyright (c) 2001 Alan Eldridge. All rights reserved.
# 
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the "Artistic License" which is 
# distributed with the software in the file LICENSE.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the Artistic
# License for more details.
# 
# $Id$
#
# 2001/05/27 jalane@sourceforge.net
#

import os
import fcntl
import struct

import pycdda.util
import pycdda.disc

######################################################################
# CONSTANTS from /usr/include/linux/cdrom.h (2.4.2)
######################################################################

# all drivers
PAUSE	= 0x5301
RESUME 	= 0x5302
PLAYMSF	= 0x5303
PLAYTRKIND	= 0x5304
READTOCHDR	= 0x5305
READTOCENTRY	= 0x5306
STOP	= 0x5307
START	= 0x5308
EJECT	= 0x5309
VOLCTRL	= 0x530a
SUBCHNL	= 0x530b
READMODE2	= 0x530c
READMODE1	= 0x530d
READAUDIO	= 0x530e
EJECT_SW	= 0x530f
MULTISESSION	= 0x5310
GET_MCN	= 0x5311
RESET	= 0x5312
VOLREAD	= 0x5313
READRAW	= 0x5314
# uniform cdrom driver
CLOSETRAY	= 0x5319
SET_OPTIONS	= 0x5320
CLEAR_OPTIONS	= 0x5321
SELECT_SPEED	= 0x5322
SELECT_DISC	= 0x5323
MEDIA_CHANGED	= 0x5325
DRIVE_STATUS	= 0x5326
DISC_STATUS	= 0x5327
CHANGER_NSLOTS	= 0x5328
LOCKDOOR	= 0x5329
DEBUG		= 0x5330
GET_CAPABILITY = 0x53331
# changer slots
CDSL_NONE 	= 0x7ffffffe
CDSL_CURRENT	= 0x7fffffff
# drive status
CDS_NO_INFO	= 0
CDS_NO_DISC	= 1
CDS_TRAY_OPEN	= 2
CDS_DRIVE_NOT_READY = 3
CDS_DISC_OK	= 4
# disc status
CDS_AUDIO	= 100
CDS_DATA_1	= 101
CDS_DATA_2	= 102
CDS_XA_2_1	= 103
CDS_XA_2_2	= 104
CDS_MIXED	= 105
# address types
CDROM_LBA	= 0x01
CDROM_MSF	= 0x02
# data track bit
CDROM_DATA_TRACK = 0x04
# leadout track
CDROM_LEADOUT	= 0xAA
# audio states (from scsi-2)
AUDIO_INVALID	= 0x00
AUDIO_PLAY	= 0x11
AUDIO_PAUSED	= 0x12
AUDIO_COMPLETED	= 0x13
AUDIO_ERROR	= 0x14
AUDIO_NO_STATUS	= 0x15

############################################################
# UTILITY FUNCTIONS
############################################################

# read subchannel data

def readsubchnl(fd):
    "Read subchannel."
    subchnl_fmt = "5B3x3Bx3Bx"
    bin = struct.pack(subchnl_fmt,
                      CDROM_MSF, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
    try:
        bin = fcntl.ioctl(fd, SUBCHNL, bin)
    except:
        return None
    return struct.unpack(subchnl_fmt, bin)

# read TOC header

def readtochdr(fd):
    "Read TOC Header."
    tochdr_fmt = "2B"
    bin = struct.pack(tochdr_fmt, 0, 0)
    bin = fcntl.ioctl(fd, READTOCHDR, bin)
    if not bin:
        return None
    return struct.unpack(tochdr_fmt, bin)

# read 1 TOC entry (index is 1 based)

def readtocentry(fd, i):
    "Read 1 TOC Entry"
    tocentry_fmt = "3Bx3BxB3x"
    bin = struct.pack(tocentry_fmt,
                      i, 0, CDROM_MSF, 0, 0, 0, 0)
    try:
        bin = fcntl.ioctl(fd, READTOCENTRY, bin)
    except:
        return None
    trk = struct.unpack(tocentry_fmt, bin)
    tinfo = pycdda.disc.track()
    tinfo.indx = trk[0]
    tinfo.ctl = trk[1] & 0xf
    tinfo.adr = (trk[1] >> 4) & 0xf
    tinfo.fmt = trk[2]
    tinfo.locMSF = tuple(trk[3:6])
    tinfo.locFrm = pycdda.util.msf_to_frm(tinfo.locMSF)
    tinfo.dmode = trk[6]
    if tinfo.ctl & CDROM_DATA_TRACK:
        tinfo.type = "data"
    else:
        tinfo.type = "audio"
    return tinfo

# fill in a disc object

def get_discinfo(fd, get_toc = 0):
    "Get current disc info."
    dinfo = pycdda.disc.info()
    # drive status
    try:
        n = fcntl.ioctl(fd, DRIVE_STATUS, CDSL_CURRENT)
    except:
        return None
    if n >= 0 and n != CDS_NO_INFO and n != CDS_DISC_OK:
        return None
    # read subchannel
    subchnl = readsubchnl(fd)
    if not subchnl:
        return None
    # check audio status
    if subchnl[1] and \
       (subchnl[1] < AUDIO_INVALID or subchnl[1] > AUDIO_NO_STATUS):
        return None
    if subchnl[1] == AUDIO_PLAY:
        dinfo.mode = "play"
    elif subchnl[1] == AUDIO_PAUSED:
        dinfo.mode = "pause"
    elif subchnl[1] == AUDIO_COMPLETED:
        dinfo.mode = "done"
    else:
        dinfo.mode = ""
    # get current locn on disc
    dinfo.locMSF = tuple(subchnl[5:8])
    dinfo.locFrm = pycdda.util.msf_to_frm(dinfo.locMSF)
    if not get_toc:
        dinfo.lenMFS = None
        dinfo.lenFrm = None
        dinfo.curTrk = None
        dinfo.disctoc = None
        return dinfo
    # get TOC hdr
    tochdr = readtochdr(fd)
    if not tochdr:
        return None
    dtoc = pycdda.disc.TOC()
    dtoc.tochdr = tochdr
    dtoc.maxTrk = tochdr[1]
    dtoc.rgTrk = []
    # read TOC entries for tracks
    for i in range(0, dtoc.maxTrk):
        tinfo = readtocentry(fd, i+1)
        if not tinfo:
            return None
        dtoc.rgTrk.append(tinfo)
    # get leadout track
    tinfo = readtocentry(fd, CDROM_LEADOUT)
    if not tinfo:
        return None
    dtoc.rgTrk.append(tinfo)
    dinfo.lenMSF = tinfo.locMSF
    dinfo.lenFrm = tinfo.locFrm
    # calc track lengths
    # same rounding algorithm as grip uses
    for i in range(0, len(dtoc.rgTrk)):
        tinfo = dtoc.rgTrk[i]
        if i > 0:
            tlast = dtoc.rgTrk[i-1]
            tlast.lenFrm = tinfo.locFrm - tlast.locFrm
            tlast.lenMSF = pycdda.util.frm_to_msf(tlast.lenFrm)
            tlast.lenSec = pycdda.util.frm_to_sec(tinfo.locFrm) \
                           - pycdda.util.frm_to_sec(tlast.locFrm)
            tlast.lenMinSec = (tlast.lenSec / 60, tlast.lenSec % 60)
    tinfo.lenFrm = 0
    tinfo.lenMSF = pycdda.util.frm_to_msf(tinfo.lenFrm)
    tinfo.lenSec = 0
    tinfo.lenMinSec = (tinfo.lenSec / 60, tinfo.lenSec % 60)    
    # set up mapping for track #s
    dtoc.mpTrk = {}
    for i in dtoc.rgTrk:
        dtoc.mpTrk[i.indx] = i
    # find current track
    tpos = pycdda.disc.pos()
    tpos.iTrk = 0
    for tinfo in dtoc.rgTrk:
        if tinfo.locFrm > dinfo.locFrm:
            break
        tpos.iTrk = tpos.iTrk + 1
    # find locn within current track
    tinfo = dtoc.rgTrk[ tpos.iTrk ]
    tpos.trkFrm = dinfo.locFrm - tinfo.locFrm
    if (tpos.trkFrm < 0):
        tpos.trkFrm = 0
    tpos.trkMSF = pycdda.util.frm_to_msf(tpos.trkFrm)
    # that's all, folks
    dinfo.toc = dtoc
    dinfo.curTrk = tpos
    return dinfo

# play a range of audio frames

def play_frames(fd, rng):
    """Play a range of frames on the disc.

    If end == 0, then play to end of disc.
    Returns a pair (beg,end) if successful.
    """
    playmsf_fmt = "6B"
    beg = rng[0]
    end = rng[1]
    if end <= beg:
        return None
    tinfo = readtocentry(fd, 1)
    if not tinfo:
        return None
    # some discs ...
    if beg < tinfo.locFrm:
        #beg = tinfo.locFrm
        beg = tinfo.locFrm + 1 
    tinfo = readtocentry(fd, CDROM_LEADOUT)
    if not tinfo:
        return None
    elif not end or end >= tinfo.locFrm:
        end = tinfo.locFrm - 1
    begMSF = pycdda.util.frm_to_msf(beg)
    endMSF = pycdda.util.frm_to_msf(end)
    try:
        fcntl.ioctl(fd, START, 1)
    except:
        return None
    bin = struct.pack(playmsf_fmt,
                      begMSF[0], begMSF[1], begMSF[2],
                      endMSF[0], endMSF[1], endMSF[2])
    try:
        bin = fcntl.ioctl(fd, PLAYMSF, bin)
    except:
        return None
    return (beg, end)
    
# is door (tray) open?

def tray_state(fd):
    """Get tray state."""
    try:
        n = fcntl.ioctl(fd, DRIVE_STATUS, CDSL_CURRENT)
    except:
        return None
    if n < 0:
        return None
    elif n == CDS_TRAY_OPEN:
        return "open"
    else:
        return "closed"

############################################################
# impl
#   CDROM device implementation class
############################################################

class cdrom:
    name = "linux"
    version = "$Id$"
    def __init__(self, fd):
        self.fd = fd
    def __del__(self):
        os.close(self.fd)
    def reset(self):
        """Hard reset"""
        try:
            return fcntl.ioctl(self.fd, RESET, 1)
        except:
            return None
    def debug(self, on = 1):
        """Driver debugging"""
        try:
            return fcntl.ioctl(self.fd, DEBUG, on)
        except:
            return None
    def get_discinfo(self, get_toc = 0):
        """Get info aout current disc.

        If there is no disc, return None.
        If get_toc is set, get the disc's
        table of contents.
        """
        return get_discinfo(self.fd, get_toc)
    def play_frames(self, rng):
        """Play a range of frames."""
        return play_frames(self.fd, rng)
    def start(self):
        """Start the drive."""
        try:
            return fcntl.ioctl(self.fd, START, 1)
        except:
            return None
    def stop(self):
        """Stop the drive."""
        try:
            return fcntl.ioctl(self.fd, STOP, 1)
        except:
            return None
    def pause(self):
        """Pause the drive."""
        try:
            return fcntl.ioctl(self.fd, PAUSE, 1)
        except:
            return None
    def resume(self):
        """Unpause the drive."""
        try:
            return fcntl.ioctl(self.fd, RESUME, 1)
        except:
            return None
    def tray_state(self):
        """Get tray state."""
        return tray_state(self.fd)
    def lock_door(self, flg = 1):
        """Lock (unlock) the door."""
        try:
            return fcntl.ioctl(self.fd, LOCKDOOR, flg)
        except:
            return None
    def open_tray(self):
        """Eject the CD tray."""
        self.lock_door(0)
        try:
            return fcntl.ioctl(self.fd, EJECT, 1)
        except:
            return None
    def close_tray(self):
        """Close the CD door."""
        try:
            return fcntl.ioctl(self.fd, CLOSETRAY, 1)
        except:
            return None

#
#EOF
##
